// $Id: StringUtil.java,v 1.1.1.1 2002/11/23 17:49:48 caron Exp $
/*
 * Copyright 1997-2000 Unidata Program Center/University Corporation for
 * Atmospheric Research, P.O. Box 3000, Boulder, CO 80307,
 * support@unidata.ucar.edu.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package thredds.util;

/**
 * String utilities
 * @author John Caron
 * @version $Id: StringUtil.java,v 1.1.1.1 2002/11/23 17:49:48 caron Exp $
 */

public class StringUtil {

  /**
   * Collapse continuous whitespace into one single " "
   */
  static public String collapseWhitespace( String s) {
    int len = s.length();
    StringBuffer b = new StringBuffer( len);
    for (int i=0; i< len; i++) {
      char c = s.charAt(i);
      if (!Character.isWhitespace(c))
        b.append(c);
      else {
        b.append(' ');
        while ((i+1<len) && Character.isWhitespace(s.charAt(i+1)))
          i++;  /// skip further whitespace
      }
    }
    return b.toString();
  }

  /**
   * Remove all whitespace in the string
   */
  static public String removeWhitespace( String s) {
    int len = s.length();
    StringBuffer b = new StringBuffer( len);
    for (int i=0; i< len; i++) {
      char c = s.charAt(i);
      if (!Character.isWhitespace(c))
        b.append(c);
    }
    return b.toString();
  }

  /**
   * Remove all occurrences of the substring sub in the string s
   */
  static public String remove( String s, String sub) {
    int len = sub.length();
    int pos;
    while (0 <= (pos = s.indexOf(sub)))
      s = s.substring(0, pos) + s.substring(pos+len);
    return s;
  }

  /**
   * Remove all occurrences of the character c in the string s
   */
  static public String remove( String s, int c) {
    if (0 > s.indexOf(c)) // none
      return s;

    StringBuffer buff = new StringBuffer(s);
    int i=0;
    while (i<buff.length()) {
      if (buff.charAt(i) == c)
        buff.deleteCharAt(i);
      else
        i++;
    }
    return buff.toString();
  }

  /**
   * Find all occurences of match strings in original, and substitute the corresponding
   *   subst string.
   * @param original: starting string
   * @param match: array of strings to match
   * @param subst: array of strings to substitute
   * @return a new string with substitutions
   */
   static public String substitute( String original, String[] match, String[] subst) {
    String s = original;
    int pos;
    for (int i=0; i< match.length; i++) {
      while (0 <= (pos = s.indexOf( match[i]))) {
        StringBuffer sb = new StringBuffer( s);
        s = sb.replace( pos, pos + match[i].length(), subst[i]).toString();
      }
    }
    return s;
  }

  /**
   * Find all occurences of the "match" in original, and substitute the "subst" string
   * @param String original: starting string
   * @param match: string to match
   * @param subst: string to substitute
   * @return a new string with substitutions
   */
   static public String substitute( String original, String match, String subst) {
    String s = original;
    int pos;
    while (0 <= (pos = s.indexOf( match))) {
      StringBuffer sb = new StringBuffer( s);
      s = sb.replace( pos, pos + match.length(), subst).toString();
    }

    return s;
  }

  /**
   * Return true if all characters are numeric
   */
  static public boolean isDigits( String s) {
    for (int i=0; i< s.length(); i++)
      if (!Character.isDigit(s.charAt(i)))
        return false;
    return true;
  }

  public static void testMatch() {

    String[] match = {"$test", "$subst", "barf"};
    String[] subst = {"TESTING123", "ok", "BARFOLA"};

    System.out.println(substitute("$test $test $test", match, subst));
    System.out.println(substitute("test $test test", match, subst));
    System.out.println(substitute("$subst $SUbST $barf", match, subst));
    System.out.println(substitute("why cant you just $test??", match, subst));

  }

  public static void main(String args[]) {

    System.out.println("<"+remove("startRE:RE:RE:end", "RE:")+"> should be <startend>");
    System.out.println("<"+remove("RE:RE:RE:end", "RE:")+"> should be <end>");
    System.out.println("<"+remove("startRE:RE:RE:", "RE:")+"> should be <start>");

    System.out.println("<"+remove("RE:RE:RE:", ":RE:")+"> should be <RERE:>");
    System.out.println("<"+remove("RE:RE:RE:", "RE")+"> should be <:::>");
    System.out.println("<"+remove("RE:middleRE:", "RE:")+"> should be <middle>");

  }

}


/* Change History:
   $Log: StringUtil.java,v $
   Revision 1.1.1.1  2002/11/23 17:49:48  caron
   thredds reorg

   Revision 1.1  2002/10/18 18:21:13  caron
   thredds server

   Revision 1.1.1.1  2002/02/15 00:01:49  caron
   import sources

*/